package com.abcseo.ws2300;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
// import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;
import java.util.Vector;

import org.apache.log4j.Logger;



/**
 * Pattern: singleton
 * <p>
 * Maintain Open2300j properties. Properties are held in the current working
 * directory in a file called open2300j.properties. This is hard coded.
 * 
 * @author <a href="mailto:david.george@gmail.com">David George</a>
 * @version $Id: $
 *          <p>
 *          Copyright  7th July 2005, David George
 *          <p>
 *          This program is published under the GNU General Public license
 * Note: uses Java 1.5 Generic Type features
 */
public class Configuration {
	public boolean shutdown = false;
	private static volatile Configuration instance = null;
	private static final String propertyFile = "OpenJ2300.properties";
	Properties props;
	private Logger logger = Logger.getLogger("Configuration");
	private Vector<WSThread> threadList = new Vector<WSThread>();

	/**
	 * Read propertyfile to configure server. If this fails we exit as we cannot
	 * continue processing without our configuration information.
	 * 
	 * properties are: WebServerPort 0 - don't run SerialPort
	 * 
	 * FTP Server FTP username FTP password FTP remote directory
	 * 
	 * RDBMS Properties ?
	 * 
	 * publish interval - interval at which data is published - either via Web
	 * or FTP update interval - interval at which data will be read from the WS
	 * 
	 * Template DocRoot - specify a directory with velocity templates to
	 * translate
	 */
	private Configuration() {

		props = new Properties();
		try {
			InputStream propFileStream = new FileInputStream(propertyFile);
			props.load(propFileStream);
		} catch (FileNotFoundException e) {
			System.err.println("Can't find property file: " + propertyFile);
			System.exit(-1);
		} catch (IOException e) {
			System.err.println("Can't load property file: " + propertyFile);
			System.exit(-1);
		}
	}
	
	/**
	 * Update period in minutes for internal webserver pages
	 */
	public int getAltitude() {
		return getInt("ws.altitude");
	}
	
	public String getLogConfigName() {
		return (String) props.get("log.config");

	}


	public String getSerialPortName() {
		return (String) props.get("station.port");

	}

	
	public long getRunSeparation()
	{
	    return Long.parseLong( (String)props.get("run.separation") );
	}
	
	   public String getIpAddress() {
	        return (String) props.get("station.ip_address");

	    }
	   
	    public String getIpPort() {
	        return (String) props.get("station.ip_port");

	    }
	
	public static final int CONNMODE_SERIAL = 1;
	public static final int CONNMODE_TCPIP = 2;
    
	
	public int  getConnMode()
	{
	    String mode = (String)props.get("station.connmode");
	    if(  ("tcpip").equalsIgnoreCase(mode) )
	        return CONNMODE_TCPIP;
	    else
	        return CONNMODE_SERIAL;
	}
	
	
	
	
	public double getRainDelta()
		throws Exception
	{
		String pr = (String) props.get("station.rain_delta");
		if( pr == null )
			return 0.0;
		else
			return Double.parseDouble( pr );
	}


	// Database configuration
	public String getJDBCDriver() {
		return (String) props.get("jdbc.driver");
	}

	public String getJDBCDatabase() {
		return (String) props.get("jdbc.database");
	}

	public String getJDBCUsername() {
		return (String) props.get("jdbc.username");
	}

	public String getJDBCPassword() {
		return (String) props.get("jdbc.password");
	}


	/**
	 * Delay in seconds between talking to weather station, minimum is 8 seconds
	 */
	public int getTemperatureRefresh() {
		return getInt("station.temperature.update") * 1000;
	}
	
	/**
	 * Delay in seconds between talking to weather station to get pressure data, minimum is 8 seconds
	 */
	public int getPressureRefresh() {
		return getInt("station.pressure.update") * 1000;
	}
	
	/**
	 * Delay in seconds between talking to weather station to get wind data, minimum is 8 seconds
	 */
	public int getWindRefresh() {
		return getInt("station.wind.update") * 1000;
	}

	/**
	 * Delay in seconds between talking to weather station to get wind data, minimum is 8 seconds
	 */
	public int getRainRefresh() {
		return getInt("station.rain.update") * 1000;
	}
	
	
	/**
	 * Update period in minutes for external ftp server pages
	 */
	public int getWeatherStationAltitude() {
		return getInt("ws.altitude");
	}
	


	private int getInt(String p) {
		int i = 0;

		try {
			i = Integer.parseInt((String) props.get(p));

		} catch (java.lang.NumberFormatException e) {
			logger.warn( "no value " + p + " set in property file");
		}

		return i;
	}

	/**
	 * Method getInstance, returns a singleton instance of this class. It is
	 * synchronized because more than one thread can make this request at once
	 * and this would confuse the loading of the configuration file
	 * 
	 * @return Configuration
	 */
	public static synchronized Configuration getInstance() {
		if (instance == null) {
			instance = new Configuration();
		}
		return instance;
	}

	/**
	 * Output document to supplied fileaname
	 */
	/*
	public void store() throws FileNotFoundException, IOException {
		FileOutputStream output = new FileOutputStream(propertyFile);
	}
	*/
	
	public void addThread(WSThread t) {
		threadList.add(t);
	}
	
	public Vector<WSThread> getThreadList() {
		return threadList;
	}
}