/*
 * Created on 12-May-2005
 *
 */
package com.abcseo.ws2300;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import org.apache.log4j.Logger;

/**
 * Set of methods related to temperature * Temperature is 4 digits e.g.: 17.3 C
 * held in two bytes BCD format + an offset of 30. e.g. 47 30
 * 
 * @author david
 * 
 */
public class Pressure extends Thread implements WSThread {
	WS2300 weatherStation;
	// private static int RELATIVE = 0x5E2;
	private static int ABSOLUTE = 0x5D8;
	double pressure;
	long time;
	private Logger logger = Logger.getLogger("Pressure");
	Configuration config;
	boolean running;
	int refreshTime;
	Connection c;
	PreparedStatement presIns, presSel;
	int altitude;

	private Pressure() {
	};

	/**
	 * Default constructor, SI (metric) units are used
	 * 
	 * @param weatherStation
	 */
	public Pressure(WS2300 weatherStation) {
		this.weatherStation = weatherStation;

		config = Configuration.getInstance();
		refreshTime = config.getPressureRefresh();
		altitude = config.getAltitude();

		try {
			logger.info("connecting to database...");
			c = DriverManager.getConnection(config.getJDBCDatabase(), config
					.getJDBCUsername(), config.getJDBCPassword());
			presIns = c.prepareStatement("INSERT INTO PRESSURE VALUES (?,?)");
			presSel = c
					.prepareStatement("SELECT * FROM PRESSURE WHERE TIME > ?");

			logger.info("... done.");
			
			// initialise object with most recent values from database
			Statement s = c.createStatement();
			s.execute("SELECT * FROM PRESSURE ORDER BY TIME DESC LIMIT 1");
			ResultSet rs = s.getResultSet();

			if (rs.next()) {
				pressure = rs.getDouble(1);
				time = rs.getTimestamp(2).getTime();
			} else {
				pressure = 0.0;
				time = 0; // the epoch
			}
		} catch (SQLException e) {
			logger.fatal(
					"Critical Database error while initialising Pressure Object "
							+ e);
			System.exit(-1);
		}
	}

	/**
	 * Sea level pressure is the pressure value obtained by the theoretical
	 * reduction or increase of station pressure to sea level. It is used to
	 * compare pressure values obtained from stations at different heights.
	 * 
	 * This is a fairly simply algorithm that gives reasonable results but
	 * doesn't into account the average temperature as does the ASOS algorithm
	 */
	private double seaLevelPressure(double p) {
		double h = (double) altitude / 1000;
		return (p / Math.exp(-0.119 * h - (0.001313 * h * h)));
	}

	public double getRelative() throws WSException {
		return pressure;
	}

	/**
	 * Update the current wind speed and direction values from the weather
	 * station
	 * 
	 * @author David George
	 * @author Open2300
	 */
	private void update(long time) throws WSException, SQLException {
		int data[];

		data = weatherStation.read(ABSOLUTE, 3);

		double p = (((data[2] & 0xF) * 1000 + (data[1] >> 4) * 100
				+ (data[1] & 0xF) * 10 + (data[0] >> 4) + (data[0] & 0xF) / 10.0));

		logger.debug( "> pressure = " + p );
		
		/*
		 * record only changes, if the figure changes then we need to look at
		 * the last figure, if this was the further back then the update time
		 * insert a record t - update with the previous figure into the database
		 * first... same for wind and temperature - this probably wants to go
		 * into update method
		 */
		if (p != pressure) {
			pressure = seaLevelPressure(p);
			update(p, time);
		}
		logger.debug( "Total pressure is: " + pressure + " hPa");
		return;
	}

	/**
	 * Insert a windspeed record into the database. This does not set the
	 * current time.
	 * 
	 * @param relative
	 *            air pressure
	 * @return if the record is older than the current record return false
	 */
	boolean update(double p, long recordTime) throws SQLException {

		if (recordTime < time) {
			return false;
		}

		presIns.setDouble(1, seaLevelPressure(p));
		presIns.setTimestamp(2, new java.sql.Timestamp(recordTime));
		presIns.execute();

		return true;
	}

	/**
	 * Main collector thread, talks to the weather station and writes data to an
	 * RDBMS
	 * 
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		running = true;
		logger.info( "Starting Pressure Collection Thread");

		while (running) {
			try {
				update(weatherStation.getCurrentTimeMillis());
			} catch (SQLException e) {
				logger.fatal( "error writing pressere to database "
						+ e);
			} catch (WSException wse) {
				logger.fatal(
						"error reading pressure from weather station " + wse);
				if( wse.shouldExit ) break;
			}

			try {
				sleep(refreshTime);
			} catch (InterruptedException e) {}
		} // while forever
		logger.info( "Stopping Rainfall Collection Thread");
	}

	public void shutdown() {
		running = false;
		this.interrupt();

		try {
			presIns.close();
			presSel.close();
			/*
			Statement stmt = c.createStatement();
			*/
			c.close();
		} catch (SQLException e) {
			logger.warn( "error closing connection to Rain table "
					+ e);
		}
	}

	public void getRelativeMinMax() {

	}

	public double getAbsolute() {
		return 0.0;
	}

	public void getAbsoluteMinMax() {

	}

	public boolean reset() {
		return true;
	}

	public double getCorrection() {
		return 0.0;
	}

	public void getTendencyForecast() {

	}
}