/*
 * Created on 12-May-2005
 *
 */
package com.abcseo.ws2300;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import org.apache.log4j.Logger;


/**
 * Set of methods related to temperature * Temperature is 4 digits e.g.: 17.3 C
 * held in two bytes BCD format + an offset of 30. e.g. 47 30
 * 
 * @author david
 * 
 * Changes
 * 3 May 2006: Petr Brouzda
 * Fix to SQL Select for Rain values
 */
public class Rain extends Thread implements WSThread {
	WS2300 weatherStation;
	public static int RAIN_TOTAL = 0x4D2;
	double totalRainfall;

	boolean SI;
	long time;
	private Logger logger = Logger.getLogger("Rain");
	Configuration config;
	boolean running;
	int refreshTime;
	private boolean firstUpdate;
	Connection c;
	PreparedStatement rainIns, rainSel;

	private Rain() {
	};

	/**
	 * Default constructor, SI (metric) units are used
	 * 
	 * @param weatherStation
	 */
	public Rain(WS2300 weatherStation) {
		firstUpdate = true;
		this.weatherStation = weatherStation;
		config = Configuration.getInstance();
		refreshTime = config.getRainRefresh();
		try {
			logger.info("connecting to database...");
			c = DriverManager.getConnection(config.getJDBCDatabase(), config
					.getJDBCUsername(), config.getJDBCPassword());
			rainIns = c.prepareStatement("INSERT INTO RAIN VALUES (?,?)");
			rainSel = c.prepareStatement("SELECT * FROM RAIN WHERE TIME > ? ORDER BY TIME ASC");
			
			logger.info("... done.");
			
			// initialise object with most recent values from database
			Statement s = c.createStatement();
			s.execute("SELECT * FROM RAIN ORDER BY TIME DESC  LIMIT 1");
			ResultSet rs = s.getResultSet();

			if (rs.next()) {
				totalRainfall = rs.getDouble(1);
				time = rs.getTimestamp(2).getTime();
			} else {
				totalRainfall = 0.0;
				time = 0; // the epoch
			}
		} catch (SQLException e) {
			logger.fatal(
					"Critical Database error while initialising Rain Object "
							+ e);
			System.exit(-1);
		}
	}

	/*
	 * 
	 * Returns: rain (double) converted to unit given in config
	 */
	public double getTotal() {
		return totalRainfall;
	}

	/**
	 * Update the current wind speed and direction values from the weather
	 * station
	 * 
	 * @author David George
	 * @author Open2300
	 */
	private void update(long time) throws WSException, SQLException, Exception {
		int data[];
		data = weatherStation.read(RAIN_TOTAL, 3);

		double r = (((data[2] >> 4) * 1000 + (data[2] & 0xF) * 100
				+ (data[1] >> 4) * 10 + (data[1] & 0xF) + (data[0] >> 4) / 10.0 + (data[0] & 0xF) / 100.0));

		logger.debug( "> rain  = " + r );
		
		if( firstUpdate )
		{
			double delta = r - weatherStation.lastHistoryRainfall;
			if( delta > 0.1 )
			{
				logger.warn( "?? Delta between real-time and history rainfall data is " + delta );
				logger.warn( "Add " + delta + " to station.rain_delta to compensate this difference." );
			}
		}
		
		/*
		 * record only changes, if the figure changes then we need to look at
		 * the last figure, if this was the further back then the update time
		 * insert a record t - update with the previous figure into the database
		 * first... same for wind and temperature - this probably wants to go
		 * into update method
		 */
		if (r != totalRainfall) {
			totalRainfall = r;
			update(totalRainfall, time);
		}
		
		logger.debug( "Total fainfall is: " + totalRainfall + " mm\n" );
		return;
	}

	/**
	 * Insert a windspeed record into the database. This does not set the
	 * current time.
	 * 
	 * @return if the record is older than the current record return false
	 */
	boolean update(double rainfall, long recordTime) throws SQLException {

		if (recordTime < time) {
			return false;
		}

		rainIns.setDouble(1, rainfall);
		rainIns.setTimestamp(2, new java.sql.Timestamp(recordTime));
		rainIns.execute();

		return true;
	}

	public boolean resetTotal() {
		return true;

	}

	/**
	 * Main collector thread, talks to the weather station and writes data to an
	 * RDBMS
	 * 
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		running = true;
		logger.info( "Starting Rainfall Collection Thread");

		while (running) {
			try {
				update(weatherStation.getCurrentTimeMillis());
			} catch (SQLException e) {
				logger.fatal( "error writing rainfall to database "
						+ e);
			} catch (WSException wse) {
				logger.fatal(
						"error reading rainfall from weather station " + wse);
				if( wse.shouldExit ) break;
			}
			catch (Exception e) {
				logger.fatal(
						"exception... ", e);
			}

			try {
				sleep(refreshTime);
			} catch (InterruptedException e) {
			}
		} // while forever
		logger.info( "Stopping Rainfall Collection Thread");
	}

	public void shutdown() {
		running = false;
		this.interrupt();

		try {
			rainIns.close();
			rainSel.close();
			/* why is it here?
			 Statement stmt = c.createStatement();
			*/
			 c.close();
		} catch (SQLException e) {
			logger.warn( "error closing connection to Rain table "
					+ e);
		}
	}


	/**
	 * Round a double value to a specified number of decimal places.
	 * 
	 * @param val
	 *            the value to be rounded.
	 * @param places
	 *            the number of decimal places to round to.
	 * @return val rounded to places decimal places.
	 */
	public double round(double val, int places) {
		long factor = (long) Math.pow(10, places);

		// Shift the decimal the correct number of places
		// to the right.
		val = val * factor;

		// Round to the nearest integer.
		long tmp = Math.round(val);

		// Shift the decimal the correct number of places
		// back to the left.
		return (double) tmp / factor;
	}
}