package com.abcseo.ws2300;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
// import java.util.Vector;
import org.apache.log4j.Logger;


/**
 * Set of methods related to temperature measurements on the WS23** range of
 * thermometers temperature is 4 digits e.g.: 17.3 C held in two bytes BCD
 * format + an offset of 30. e.g. 47 30
 * 
 * All temperatures are in centigrade
 * 
 * @author david george
 * 
 * (c) 12-May-2005 David George This program is published under the GNU General
 * Public license
 */
public class TemperatureHumidity extends Thread implements WSThread {
	// addresses in the WS2300
	final private static int OUTDOOR_TEMP = 0x373;
	final private static int OUTDOOR_HUMIDITY = 0x419;
	final private static int INDOOR_HUMIDITY = 0x3FB;
	final private static int DEWPOINT = 0x3CE;
	final private static int INDOOR_TEMP = 0x346;
	final private static int OUTDOOR_MINMAX = 0x378;
	final private static int ONE_DAY = 86400000;

	double outdoorTemperature;
	double average12hrTemperature;
	double indoorTemperature;
	int outdoorHumidity;
	int cloudbase;
	long time;
	int wsAltitude;

	WS2300 weatherStation;
	private Logger logger = Logger.getLogger("TempHum");
	Configuration config;
	boolean running;
	int refreshTime;
	Connection c;
	PreparedStatement odtIns, odtSel;
	String docRoot;
	// Vector<Interval> gi = new Vector<Interval>();

	/**
	 * Class representing a Max / Min temperature measurement
	 * 
	 * @author david
	 */
	public class MaxMin {
		public double temp_min;
		public double temp_max;
		public int min_minute;
		public int min_hour;
		public int min_day;
		public int min_month;
		public int min_year;
		public int max_minute;
		public int max_hour;
		public int max_day;
		public int max_month;
		public int max_year;
	}

	private TemperatureHumidity() {
	};

	/**
	 * Default constructor, SI (metric) units are used, create
	 * 
	 * @param weatherStation
	 */
	TemperatureHumidity(WS2300 weatherStation)  
	    {
		this.weatherStation = weatherStation;

		config = Configuration.getInstance();
		refreshTime = config.getTemperatureRefresh();
		wsAltitude = config.getWeatherStationAltitude();

		// create connection to the database
		try {
			Class.forName(config.getJDBCDriver());
		} catch (Exception e) {
			logger.fatal( "failed to load JDBC driver " + e);
			System.exit(-1);
		}

		try {
			logger.info("connecting to database...");
			c = DriverManager.getConnection(config.getJDBCDatabase(), config
					.getJDBCUsername(), config.getJDBCPassword());
			odtIns = c
					.prepareStatement("INSERT INTO OUTSIDE_TEMP VALUES (?,?,?)");
			odtSel = c
					.prepareStatement("SELECT * FROM OUTSIDE_TEMP WHERE TIME > ?");

			logger.info("... done.");
			
			// initialise object with most recent values from database
			Statement s = c.createStatement();
			s.execute("SELECT * FROM OUTSIDE_TEMP ORDER BY TIME DESC  LIMIT 1");
			ResultSet rs = s.getResultSet();

			/**
		     * Checks if latest Temperature record is older than 15 minutes. If it isn't, it means that 
		     * any other instance of application is already running - at least probably. Aplication 
		     * should terminate in this case - as it will be started again later.
		     * 
		     * Not needed for serial connection.
		     */
			if (rs.next()) {
				outdoorTemperature = rs.getDouble(1);
				outdoorHumidity = rs.getInt(2);
				time = rs.getTimestamp(3).getTime();
				
				if( 
				        ( time > (System.currentTimeMillis() - (1000L * config.getRunSeparation()) ) )
				        &&
				        ( config.getConnMode() == Configuration.CONNMODE_TCPIP )
				)
				{
				    logger.error("Latest record in database is newer than " + config.getRunSeparation() + " seconds ("+ rs.getTimestamp(3).toString() +")" );
				    logger.error("+ Application is already running here (probably)! " );
				    logger.error("+ Wait a minute, then try to start it again." );
				    System.exit(-1);
				}
			}
		} catch (SQLException e) {
			logger.fatal(
					"Critical Database error while initialising TemperatureHumidity Object "
							+ e);
			System.exit(-1);
		}
/*
		String s = config.getGraphInterval("temperature.interval");
		String[] intervals = s.split(",", -1);
		for (int x = 0; x < intervals.length; x++) {
			System.out.println(intervals[x]);
			String interval = intervals[x].trim();

			// last letter is h, m or d
			String unit = intervals[x].substring(interval.length() - 1);
			String length = intervals[x].substring(0, interval.length() - 1);
			Interval i = new Interval(unit, length);
			try {
				gi.add(i);
			} catch (NumberFormatException e) {
				logger.warn(
						"Can't parse interval value in properties file "
								+ length + unit);
			}

		}
*/		
	}
	
	
	
	/**
	 * Returns the current outdoor temperature
	 * 
	 * @param WS2300
	 *            A Weather Station Comms Object
	 * @return Current outdoor temperature
	 */
	public synchronized double getOutdoorTemperature() {
		return outdoorTemperature;
	}

	private synchronized void updateOutdoor(long t) throws WSException,
			SQLException {
		int data[];
		
		double it = getTemperature(INDOOR_TEMP);

		if (it < -29.9 || it > 69.9) {
			logger.warn( "outdoor temperature out of range " + it);
			return;
		}

		logger.debug( "> in temp = " + it );
				
		double ot = getTemperature(OUTDOOR_TEMP);

		if (ot < -29.9 || ot > 69.9) {
			logger.warn( "outdoor temperature out of range " + ot);
			return;
		}
		
		logger.debug( "> out temp = " + ot );
		
		data = weatherStation.read(INDOOR_HUMIDITY, 1);
		int ih = ((data[0] >> 4) * 10 + (data[0] & 0xF));
		
		if (ih < 0 || ih > 100) {
			logger.warn( "humidity is out of range " + ih);
			return;
		}
		
		logger.debug( "> in humidity = " + ih );


		data = weatherStation.read(OUTDOOR_HUMIDITY, 1);
		int h = ((data[0] >> 4) * 10 + (data[0] & 0xF));
		
		if (h < 0 || h > 100) {
			logger.warn( "humidity is out of range " + h);
			return;
		}
		
		logger.debug( "> out humidity = " + h );

		// only record changes
		if (t != outdoorTemperature || h != outdoorHumidity) {
			outdoorTemperature = ot;
			outdoorHumidity = h;
			updateOutdoor(ot, h, t);
		}
		// note we only set the current time if the insert into the db is
		// successful
		time = t;
	}

	/**
	 * Insert a temperature record into the database. This does not set the
	 * current time.
	 * 
	 * @return if the record is older than the current record return false
	 */
	boolean updateOutdoor(double temperature, int humidity, long recordTime)
			throws SQLException {

		if (recordTime < time) {
			return false;
		}
		odtIns.setDouble(1, temperature);
		odtIns.setInt(2, humidity);
		odtIns.setTimestamp(3, new java.sql.Timestamp(recordTime));
		odtIns.execute();

		logger.debug( "Outdoor Temperature is: " + temperature
				+ " C humidity " + humidity);
		return true;
	}

	/**
	 * Read indoor relative humidity, current value only
	 * 
	 * @return current humidity as a %
	 */
	public int getOutdoorHumidity() {
		return outdoorHumidity;
	}

	/**
	 * Read indoor min/max temperatures with timestamps
	 * 
	 * @return MinMax temperature object
	 */
	public MaxMin getIndoorMinMax() {

		return null;
	}

	/**
	 * Read outdoor min/max temperatures with timestamps
	 * 
	 * @return MinMax temperature object
	 */
	public MaxMin getOutdoorMaxMin() {
		return getMaxMinTemperature(OUTDOOR_MINMAX);
	}

	/**
	 * Reset indoor min/max temperatures
	 */
	public boolean indoorReset() {
		return true;
	}

	/**
	 * Reset outdoor min/max temperatures
	 * 
	 */
	public boolean outdoorReset() {
		return true;
	}

	public double getDewpoint() {
		return calculateDewpoint();
	}

	/**
	 * Calculates the current dewpoint temperature
	 * 
	 * @link http://www.faqs.org/faqs/meteorology/temp-dewpoint/
	 * @author Open2300
	 * @return Current dewpoint value
	 */
	private double calculateDewpoint() {
		double A, B, C; // Intermediate values used for dewpoint calculation

		A = 17.2694;
		B = (outdoorTemperature > 0) ? 237.3 : 265.5;
		C = (A * outdoorTemperature) / (B + outdoorTemperature)
				+ Math.log((double) outdoorHumidity / 100);

		return B * C / (A - C);
	}

	/**
	 * Read dewpoint min/max temperatures with timestamps
	 */
	public MaxMin getDewpointMinMax() {
		return null;
	}

	/**
	 * Reset outdoor min/max temperatures with timestamps
	 * 
	 */
	public boolean dewpointReset() {
		return true;
	}

	/**
	 * Returns the current indoor temperature
	 * 
	 * @param WS2300
	 *            A Weather Station Comms Object
	 * @return Current outdoor temperature
	 */
	public double getIndoor() throws WSException {
		indoorTemperature = getTemperature(INDOOR_TEMP);
		return indoorTemperature;
	}

	/**
	 * Returns the Cloudbase
	 * 
	 * Consider a parcel (volume) of air at ground elevation z, characterized by
	 * its temperature T0 and its dew point Tdo. The dew point is the
	 * temperature to which this parcel of air would have to be cooled, without
	 * changing its pressure or moisture content, in order to condense some of
	 * the vapor into tiny droplets, that is, to initiate the formation of fog.
	 * 
	 * When the temperature and dew point are equal (T=Td), the relative
	 * humidity is 100 percent.
	 * 
	 * As a parcel of air is lifted, it expands and cools at a constant, known
	 * rate G, namely, 9.8 degrees Celsius per kilometer. The temperature T of
	 * the parcel at some distance above the ground is therefore given by the
	 * simple formula
	 * 
	 * T = ToG(zzo)
	 * 
	 * where z is the altitude of the parcel. Though the moisture content of the
	 * parcel remains the same as the parcel cools, the pressure drops as the
	 * parcel rises. Hence we might expect the dew point to change. In fact, the
	 * dew point of the rising parcel decreases at a nearly constant rate Gdew,
	 * but not as quickly as the temperature. At altitude z, the dew point of
	 * the parcel is approximately
	 * 
	 * Td = TdoGdew (zzo).
	 * 
	 * If the starting parcel temperature is higher than the dew point, and the
	 * temperature decreases more rapidly than dew point during lifting, then at
	 * some altitude zLCL, the temperature and dew point become equal, the
	 * relative humidity reaches 100 percent, and vapor begins to condense. This
	 * level is called, appropriately, the lifting-condensation level (LCL)
	 * because this is where a cloud begins to form within a lifted parcel.
	 * Above cloud base, the parcel temperature decreases less rapidly than 9.8
	 * degrees/km because condensation warms the parcel internally.
	 * 
	 * At the LCL, z=zLCL and T=Td. If we make these substitutions in both
	 * equations and then subtract the bottom equation from the top equation,
	 * the result is 0 = (ToTdo)  (zLCLzo)(GGdew)
	 * 
	 * or
	 * 
	 * zLCLzo =
	 * 
	 * The left hand side is the height above ground of the base of the clouds.
	 * We have noted already that G is 9.8degrees Celsius/km. Bohren and
	 * Albrecht show that Gdew is approximately 1.8degrees Celsius/km. The
	 * denominator is thus 8.0degrees Celsius/km. If we enter the dew point
	 * depression (ToTdo) in degrees Celsius, we find the height of the cloud
	 * base above ground in kilometers. If we express the lapse rates in degrees
	 * Celsius per meter, however, the denominator becomes 0.008, and the
	 * equation may be rewritten
	 * 
	 * zLCLzo = 125 (ToTdo).
	 * 
	 * Now the height of the cloud base comes out in meters, and we have
	 * demonstrated the simple formula that was the basis of your question.
	 * 
	 * The units of the constant (125) are meters/degrees Celsius. If we change
	 * the units to feet/ degrees F, the numerical value of the constant becomes
	 * 228. Thus, the height of the cloud base in feet is equal to the surface
	 * dew point depression expressed in degrees Fahrenheit multiplied by 228,
	 * not 222.
	 * 
	 * The condition under which the formula holds is rather specialized. The
	 * atmosphere between the ground and cloud base must be well mixed; that is,
	 * the actual decrease in temperature with height, the environmental lapse
	 * rate, must match the parcels decrease in temperature as it is lifted,
	 * namely, 9.8degrees Celsius/km. As long as the parcel and its surroundings
	 * have the same temperature, vertical motion is freely permitted and the
	 * air will be well mixed. If the environmental lapse rate is less than
	 * 9.8degrees Celsius/km, the rising parcel quickly becomes cooler than its
	 * surroundings (negatively buoyant) and tends to sink back toward its
	 * starting point. Vertical motion is inhibited.
	 * 
	 * Well-mixed conditions occur most frequently on hot summer afternoons,
	 * especially following a sunny morning. The layer in which the lapse rate
	 * is 9.8degrees Celsius/km builds up from the ground, which is heated by
	 * the sun and transfers its warmth to the overlying air. In dry climates
	 * where the heating is intense, this layer can become 10,000 feet deep.
	 * 
	 * Because the layer is well mixed, the base of cumulus clouds that form
	 * near the top of this layer as the result of convective updrafts
	 * (thermals) are all near the same altitude. The air close to the ground is
	 * normally not well mixed at night or in the early morning. At those times,
	 * the formula does not apply. How accurate is the formula under well-mixed
	 * conditions? I calculated the dew point lapse rate dew under a variety of
	 * conditions from an equation given by Bohren and Albrecht. The values
	 * ranged from 1.6degrees Celsius/km to nearly 1.9degrees Celsius/km. This
	 * leads to values of the constant in the metric version of the equation
	 * between 122.0 and 126.5m/degrees Celsius, a variation of roughly 2
	 * percent on either side of 125. The surface dew point depression (ToTdo)
	 * probably cannot be measured with better precision. The formula is
	 * therefore accurate to within about 2 percent.
	 * 
	 * @link http://www.weatherwise.org/qr/qry.00.hicloud.html
	 */
	public int getCloudBase() {
		double dewpoint = calculateDewpoint();

		cloudbase = (int) (outdoorTemperature - dewpoint) * 125;
		if (cloudbase < 0) {
			return 0;
		}

		return cloudbase + wsAltitude;
	}

	/**
	 * Calculates the zero isotherm height
	 * 
	 * @return height
	 */
	public int getZeroIsotherm() {
		return calculateZeroIsotherm(outdoorTemperature);
	}

	private int calculateZeroIsotherm(double temperature) {
		return (int) (temperature / 0.0065) + wsAltitude;
	}

	public String getAltitudeUnits() {
		// if (SI) {
		return "meters";
		// } else {
		// return "feet";
		// }
	}

	private double getTemperature(int address) throws WSException {
		int data[];
		double temperature = 0.0;

		data = weatherStation.read(address, 2);
		temperature = ((((data[1] >> 4) * 10 + (data[1] & 0xF) + (data[0] >> 4)
				/ 10.0 + (data[0] & 0xF) / 100.0) - 30.0));

		return temperature;
	}

	/**
	 * Do something with Exception conditions.
	 * 
	 * @param address
	 * @return
	 */
	private MaxMin getMaxMinTemperature(int address) {
		int data[];
		MaxMin mm = new MaxMin();

		try {
			data = weatherStation.read(OUTDOOR_MINMAX, 15);
			mm.temp_min = ((((data[1] >> 4) * 10 + (data[1] & 0xF)
					+ (data[0] >> 4) / 10.0 + (data[0] & 0xF) / 100.0) - 30.0));
			mm.temp_max = ((data[4] & 0xF) * 10 + (data[3] >> 4)
					+ (data[3] & 0xF) / 10.0 + (data[2] >> 4) / 100.0) - 30.0;

			mm.min_minute = ((data[5] & 0xF) * 10) + (data[4] >> 4);
			mm.min_hour = ((data[6] & 0xF) * 10) + (data[5] >> 4);
			mm.min_day = ((data[7] & 0xF) * 10) + (data[6] >> 4);
			mm.min_month = ((data[8] & 0xF) * 10) + (data[7] >> 4);
			mm.min_year = 2000 + ((data[9] & 0xF) * 10) + (data[8] >> 4);

			mm.max_minute = ((data[10] & 0xF) * 10) + (data[9] >> 4);
			mm.max_hour = ((data[11] & 0xF) * 10) + (data[10] >> 4);
			mm.max_day = ((data[12] & 0xF) * 10) + (data[11] >> 4);
			mm.max_month = ((data[13] & 0xF) * 10) + (data[12] >> 4);
			mm.max_year = 2000 + ((data[14] & 0xF) * 10) + (data[13] >> 4);

			mm.temp_min = mm.temp_min * 9 / 5 + 32;
			mm.temp_max = mm.temp_max * 9 / 5 + 32;

		} catch (Exception e) {
			logger.warn( "error reading max/min temperature"
					+ e.toString());
		}

		return mm;
	}

	/**
	 * Convert centegrade to fahrenheit
	 * 
	 * @param temperature
	 */
	public static double cToF(double temperature) {
		return temperature * 9 / 5 + 32;
	}

	/*
	 * Last time date was set - maybe put in a superclass?
	 */
	public String getDate() {
		Date d = new Date(time);
		DateFormat df = new SimpleDateFormat();
		return df.format(d);
	}

	/**
	 * Main collector thread, talks to the weather station and writes data to an
	 * RDBMS
	 * 
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		running = true;
		logger.info( "Starting Temperature Collection Thread");

		while (running) {
			try {
				updateOutdoor(weatherStation.getCurrentTimeMillis());
			} catch (SQLException e) {
				logger.fatal( "error writing to database " + e);
			} catch (WSException wse) {
				logger.fatal( "error reading from weather station "
						+ wse);
				if( wse.shouldExit ) break;
			}

			try {
				sleep(refreshTime);
			} catch (InterruptedException e) {

			}
		} // while forever
		logger.info( "Stopping Temperature Collection Thread");
	}

	public void shutdown() {
		running = false;
		this.interrupt();

		try {
			odtIns.close();
			odtSel.close();
			/*
			 Statement stmt = c.createStatement();
			// only needs to be done in one thread
			stmt.executeUpdate("SHUTDOWN");
			stmt.close();
			*/
			c.close();
			
			logger.info( "Database Shutdown");
		} catch (SQLException e) {
			logger.fatal( "error shutting down database " + e);
		}
	}


}
