/*
 * Created on 10-May-2005
 *
 * 
 */
package com.abcseo.ws2300;

import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.UnsupportedCommOperationException;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;

import org.apache.log4j.xml.DOMConfigurator;
import org.apache.log4j.Logger;


/**
 * @author David George
 * 
 * This class supports basic IO operations on La Crosse WS2300 series weather
 * stations.
 * 
 * Information on the protocol comes from this page:
 * http://perso.wanadoo.fr/claude.ocquidant/autrespages/leprotocol/protocol-eng.htm
 * 
 * Some basic functionality, in part ported from Open2300:
 * open2300.sourceforge.net
 */
public class WS2300 {
	public static int HISTORY_INFO = 0x6B2;
	public static int HISTORY_RECORD = 0x6C6;
	public static int CURRENT_TIME = 0x23B;

	// various data access objects
	TemperatureHumidity temperature;
	Wind wind;
	Rain rain;
	Pressure pressure;
	InputStream inputStream = null;
	OutputStream outputStream;
	Socket sock2;
	
	SerialPort serialPort = null;
	
	private int maxRetries = 5;
	private Logger logger;
	private Logger logger2;
	public double lastHistoryRainfall;
	
	Configuration config;
	boolean SI;

	/**
	 * Set up communications with the WS2300 device. The weather station talks
	 * at 2400 baud 8 bits, no parity, one stop bit.
	 * 
	 * If the weather station is not connected only historical information from
	 * the datebase will be available.
	 */
	public WS2300() {
		config = Configuration.getInstance();
		
		
		DOMConfigurator.configure( config.getLogConfigName() );
		logger = Logger.getLogger("WS2300");
		logger2 = Logger.getLogger("data");

		// allocate data collection objects
		temperature = new TemperatureHumidity(this);
		wind = new Wind(this);
		rain = new Rain(this);
		pressure = new Pressure(this);
		
		if( config.getConnMode() == config.CONNMODE_SERIAL )
		{
		    
		    CommPortIdentifier portId = null;
		    
            // *********** serial port settings *****************

        		try {		
        			portId = CommPortIdentifier.getPortIdentifier(config
        					.getSerialPortName());
        			if (portId.getPortType() != CommPortIdentifier.PORT_SERIAL) {
        				logger.warn( config.getSerialPortName()
        						+ " is not a serial port");
        			}
        		} catch (NoSuchPortException e) {
        			logger.warn( "can't open port "
        					+ config.getSerialPortName());
        			return;
        		}
        
        		try {
        			serialPort = (SerialPort) portId.open("WS2300", 2000);
        		} catch (PortInUseException e) {
        			logger.warn( "Cannot Open " + portId.getName() + " "
        					+ e);
        			return;
        		}
        
        		try {
        		    serialPort.enableReceiveTimeout(500);
        		} catch (UnsupportedCommOperationException e) {
        			logger.warn( "receive timeout not supported" + e);
        			return;
        		}
        		try {
        			inputStream =serialPort.getInputStream();
        			outputStream = serialPort.getOutputStream();
        		} catch (IOException e) {
        			logger.warn( "Error creating serial port streams "
        							+ e);
        			serialPort.close();
        			return;
        		}
        
        		try {
        		    serialPort.setSerialPortParams(2400, SerialPort.DATABITS_8,
        					SerialPort.STOPBITS_1, SerialPort.PARITY_NONE);
        		} catch (UnsupportedCommOperationException e) {
        			logger.warn( "Could not configure serial port " + e);
        			serialPort.close();
        			return;
        		}
        	// *********** serial port settings *****************
		}
		else
		{
		    // *********** TCP/IP port settings *****************
		    
		    // potrebujeme naplnit 	    inputStream 	    outputStream
		    try
		    {
    		    sock2 = new Socket(
    		            config.getIpAddress(), 
    		            Integer.parseInt(config.getIpPort()) ); // hned se pipojme
    		    
    		    sock2.setSoTimeout(5000);
    		    sock2.setKeepAlive(true);
    		    
    		    inputStream = sock2.getInputStream();
    		    outputStream = sock2.getOutputStream();
		    } 
		    catch( Exception ee )
		    {
		        logger.fatal( "Could not configure TCP/IP port", ee);
		        return;
		    }
		    
		    // *********** TCP/IP port settings *****************
		}
		

		// weather station online get history information and populate data
		// collection objects
		try {
			logger.info("reading history...");
			HistoryInfo hinfo = readHistoryInfo();
			logger.info("... history read. " + hinfo.records + " records ready in history." );
			
			long recordTime = hinfo.getTimeMilliseconds();

			int record = hinfo.lastRecord;
			boolean doTemp = true;
			boolean doWind = true;
			boolean doRain = true;
			boolean doPressure = true;
			
			for (int i = 0; i < hinfo.records; i++) 
			{
				HistoryRecord hrecord = readHistoryRecord(record);
				Date d = new Date(recordTime);
				DateFormat df = new SimpleDateFormat();
				logger.debug( "[" + record + "] " + df.format(d) + ": "
						+ hrecord);
				
				hrecord.rainfall += config.getRainDelta();
				
				if( i == 0 )
				{
					// for solving the strange delta between realtime and history rainfall data
					lastHistoryRainfall = hrecord.rainfall;
				}

				if (doTemp) {
					doTemp = temperature.updateOutdoor(hrecord.outdoorTemperature,
						hrecord.outdoorHumidity, recordTime);
				}
				if (doWind) {
					doWind = wind.update(hrecord.windspeed, hrecord.winddir, hrecord.windspeed, recordTime);
				}
				if (doRain) {
					doRain = rain.update(hrecord.rainfall, recordTime);
				}
				if (doPressure) {
					doPressure = pressure.update(hrecord.pressure, recordTime);
				}
				if (!doTemp && !doWind && !doRain && !doPressure) {
					break;
				}
				record = hrecord.nextRecord;
				recordTime -= hinfo.getIntervalMillseconds();
				
			}// for

		} catch (Exception e) {
			e.printStackTrace();
			if( logger != null )
				logger.fatal( "Unhandled exception", e );
		}

		// temperature must be first as it also shuts down the database
		temperature.start();
		config.addThread(temperature);
		wind.start();
		config.addThread(wind);
		rain.start();
		config.addThread(rain);
		pressure.start();
		config.addThread(pressure);
	}

	/**
	 * Reset the Weather Station. This is done by sending a 0x06 command until
	 * 0x02 is received
	 */
	private synchronized void reset() {
		try {
		    logger2.debug( "reset" );
		    
		    if( config.getConnMode() == config.CONNMODE_SERIAL )
		    {
		        /*
		         http://www.lavrsen.dk/foswiki/bin/view/Open2300/OpenWSAPI
		         
		        DTR and RTS are not used for handshake. They are steady DTR at negative voltage and 
		        RTS at positive voltage.
		        Since there is no ground it seems that the WS2300 uses the DTR and RTS to define high 
		        and low. DTR must be low and RTS must be high for the communication to work.
		        Some USB-Serial adapters do not support DTR and RTS. If your WS2300 does not communicate, 
		        check the voltage of DTR and RTS at the connector.
		        */
		        
		        serialPort.setDTR(false);
		        serialPort.setRTS(true);
		    }

			outputStream.write(0x06);
			for (int i = 0; i < maxRetries; i++) {
				try {
					Thread.sleep(100);
				} catch (InterruptedException e) {
				}
				while (inputStream.available() > 0) {

					int data = inputStream.read();
					// DEBUG
					logger2.debug( "data available " + data);
					
					if (data == 2) {
						// consume rest of buffer
					} else {
						outputStream.write(0x06);
					}

				}// while
			}// for

			
		} catch (IOException e) {
			logger.warn( "Error resetting weather station " + e);
			return;
		}
	}

	/**
	 * Handles timeouts and read errors when communicating with the Weather Station
	 * 
	 * @param address
	 * @param numBytes
	 * @return
	 * @throws WSException
	 */
	public int[] read(int address, int numBytes) throws WSException {

		for (int i = 0; i < maxRetries; i++) {
			reset();
			try {
				return rawRead(address, numBytes);
			} 
			catch( SocketException ee )
			{
			    logger.error( "Socket error: ", ee );
			    throw new WSException( "Socket error: " + ee.getMessage(), true );
			}
			catch( SocketTimeoutException ee )
			{
			    logger2.debug(  "Read time out" );
			}
			catch (WSException we )
			{
				logger2.debug(  we.toString() );
				// System.out.println( "!!! " + we.toString() );
			}
			catch (Exception e) {
				// log exception
				logger.debug( "Exception ", e );
			}

			// station may be busy, incremental backoff
			try {
				Thread.sleep((i + 1) * 1000);
			} catch (InterruptedException e) {
			}
		}// for

		throw new WSException("read data error at address " + address);
	}

	/**
	 * rawRead
	 * 
	 * This routine talks directly to the weather station. It is synchronized to prevent multiple
	 * from sending data at the same time.
	 * 
	 * @param address
	 * @param numBytes
	 * @throws IOException
	 * @throws WSException
	 *             error in the communication with the weather station
	 * 
	 * The WS2300 has to be told the address to work with. Addresses are in the
	 * range: 0000-13B0. as four separate bytes. The station needs to be set
	 * into addressing mode by adding the hex value 0x82 to each value.
	 */
	private synchronized int[] rawRead(int address, int numBytes) throws IOException, WSException {
		int[] encodedAddress = encodeAddress(address);
		int[] data = new int[numBytes];
		int response;

		// DEBUG
		logger2.debug( "read adr:" + address + " len:" + numBytes);

		if (address < 0 || address > 0x13B0) {
			throw new WSException("bad address range " + address);
		}

		// we now write each byte and examine the returned checksum
		for (int i = 0; i < 4; i++) {
			outputStream.write(encodedAddress[i]);
			// read checksum
			response = inputStream.read();
			// DEBUG
			logger2.debug( "data available " + response);
			
			if (response != addressCSum(encodedAddress[i], i)) {
				throw new WSException("bad address encoding at address " + address + " at " + i);
			}

		}// for

		// number of bytes to read
		int number = encodeNumber(numBytes);
		outputStream.write(number);
		// read checksum
		response = inputStream.read();
		
		// DEBUG
		logger2.debug( "data available " + response);
		
		if (response != numBytesCSum(numBytes)) {
			throw new WSException("bad number encoding " + address);
		}

		// read the data
		for (int i = 0; i < numBytes; i++) {
			data[i] = inputStream.read();
			// DEBUG
			logger2.debug( "data[" + i + "] " + data[i]);
		}// for

		// read checksum
		response = inputStream.read();
		if (response != dataCSum(data)) {
			throw new WSException("data error " + address);
		}

		return data;
	}



	/**
	 * encodeAddress converts an integer address to the form needed by the
	 * WS-2300 when sending commands.
	 * 
	 * @param address
	 *            WS2300 address range
	 * @returns four bytes representing the encoded address
	 * 
	 * @author open2300
	 */
	private int[] encodeAddress(int address) {
		int[] encodedAddress = new int[4];

		int nibble;
		for (int i = 0; i < 4; i++) {
			nibble = (address >> (4 * (3 - i))) & 0x0F;
			encodedAddress[i] = (0x82 + (nibble * 4));
		}

		return encodedAddress;
	}

	/**
	 * numberEncoder converts the number of bytes we want to read to the form
	 * needed by the WS-2300 when sending commands.
	 * 
	 * @param number
	 *            (0-15 (nibble))
	 * @returns number in WS2300 format *
	 * @author open2300
	 */
	private int encodeNumber(int number) {
		int codedNumber;

		codedNumber = (0xC2 + number * 4);
		if (codedNumber > 0xfe) {
			codedNumber = 0xfe;
		}

		return codedNumber;
	}

	/**
	 * calculates the checksum for the addresses sent to WS2300
	 * 
	 * For each addressing byte sent an acknowledgement byte is returned. This
	 * has the sequence number (0-3) encoded in the high order four bits and
	 * address in the low order four bits
	 * 
	 * @param acknowledgement
	 * @param sequenceNumber
	 * 
	 * @returns calculated checksum
	 * 
	 */
	private int addressCSum(int acknowledgement, int sequenceNumber) {
		return (sequenceNumber << 4) + ((acknowledgement - 0x82) / 4);

	}

	/**
	 * command_check4 calculates the checksum for the last command which is sent
	 * just before data is received from WS2300
	 * 
	 * Input: number of bytes requested
	 * 
	 * Returns: expected response from requesting number of bytes
	 * 
	 * @author open2300
	 */
	private int numBytesCSum(int number) {
		return 0x30 + number;
	}

	/**
	 * data_checksum calculates the checksum for the data bytes received from
	 * the WS2300
	 * 
	 * Input: pointer to array of data to check number of bytes in array
	 * 
	 * Returns: calculated checksum as unsigned char *
	 * 
	 * @author open2300
	 */
	private int dataCSum(int[] data) {
		int checksum = 0;

		for (int i = 0; i < data.length; i++) {
			checksum += data[i];
		}// for

		checksum &= 0xFF;

		return checksum;
	}

	/**
	 * turn light on or off
	 * 
	 * @param on
	 */
	void light(boolean on) {

	}

	void close() {
	    if( config.getConnMode() == config.CONNMODE_SERIAL )
	    {
	        serialPort.close();
	    }
	    else
	    {
	        try
	        {
	            sock2.close();
	        }
	        catch( Exception ee )
	        {
	            logger.error("can't close socket", ee );
	        }
	        
	            
	    }
	    
	}

	/**
	 * @return time in milliseconds from WeatherStation
	 */
	public long getCurrentTimeMillis() {
		Calendar c = Calendar.getInstance();
		int data[];
		try {
			if ((data = read(CURRENT_TIME, 6)) != null) {

				int minute = (data[0] & 0xF) + ((data[0] >> 4) * 10);
				int hour = (data[1] & 0xF) + ((data[1] >> 4) * 10);
				int day = ((data[3] & 0xF) * 10) + (data[2] >> 4);
				int month = ((data[4] & 0xF) * 10) + (data[3] >> 4);
				int year = 2000 + ((data[5] & 0xF) * 10) + (data[4] >> 4);

				c.set(year, month - 1, day, hour, minute);
				return c.getTimeInMillis();
			}
		} catch (Exception e) {
			logger.warn( "Error talking to weather station " + e.toString());
		}

		return -1;

	}

	public class HistoryInfo {
		public int interval;
		public int countdown;
		public int minute;
		public int hour;
		public int day;
		public int month;
		public int year;
		public int records;
		public int lastRecord;

		public String toString() {
			return "interval " + interval + " minutes, next measurement "
					+ countdown + " last measurement " + day + " " + month
					+ " " + year + ", " + hour + ":" + minute + " no records "
					+ records + " last record " + lastRecord;
		}

		public long getTimeMilliseconds() {
			Calendar c = Calendar.getInstance();

			c.set(year, month - 1, day, hour, minute);
			return c.getTimeInMillis();
		}

		public long getIntervalMillseconds() {
			return interval * 60 * 1000;
		}
	}

	/**
	 * read_history_info Read the history information from the weather station.
	 * The lastRecord pointer is the last record stored in memory and is used by
	 * readHistoryRecord.
	 * 
	 * @returns HistoryInfo history information from weather station such as
	 *          last record time etc.
	 * 
	 * @author open2300
	 */
	private HistoryInfo readHistoryInfo() throws WSException {
		int data[];
		HistoryInfo hinfo = new HistoryInfo();

		data = read(HISTORY_INFO, 10);

		hinfo.interval = (data[1] & 0xF) * 256 + data[0] + 1;
		hinfo.countdown = data[2] * 16 + (data[1] >> 4) + 1;
		hinfo.minute = ((data[3] >> 4) * 10) + (data[3] & 0xF);
		hinfo.hour = ((data[4] >> 4) * 10) + (data[4] & 0xF);
		hinfo.day = ((data[5] >> 4) * 10) + (data[5] & 0xF);
		hinfo.month = ((data[6] >> 4) * 10) + (data[6] & 0xF);
		hinfo.year = 2000 + ((data[7] >> 4) * 10) + (data[7] & 0xF);
		hinfo.records = data[9];

		hinfo.lastRecord = data[8];

		return hinfo;
	}

	/**
	 * read_history_info Read the history record from weather station.
	 * 
	 * @param record -
	 *            valid record pointer calculated from readHistoryInfo()
	 * 
	 * @returns HistoryRecord weather record object
	 * 
	 * @author open2300
	 */
	private class HistoryRecord {
		public double indoorTemperature;
		public double outdoorTemperature;
		public double pressure;
		public int indoorHumidity;
		public int outdoorHumidity;
		public double rainfall;
		public double windspeed;	// m/s
		public double winddir;
		public double windchill;
		public int nextRecord;

		public String toString() {
			return "indoor temperature " + indoorTemperature
					+ ", outdoor temperature " + outdoorTemperature
					+ ", pressure " + pressure + ", indoorHumidity "
					+ indoorHumidity + ", outdoorHumidity " + outdoorHumidity
					+ ", rainfall " + rainfall + ", windspeed " + windspeed
					+ ", winddir " + winddir;
		}
	}

	private HistoryRecord readHistoryRecord(int record) throws WSException {
		int data[];
		HistoryRecord hrecord = new HistoryRecord();
		//int bytes = 10;
		int tempint;

		//double wind_kmph;

		int address = HISTORY_RECORD + (record * 19);
		data = read(address, 10);

		tempint = (data[4] << 12) + (data[3] << 4) + (data[2] >> 4);

		hrecord.pressure = 1000 + (tempint % 10000) / 10.0;

		if (hrecord.pressure >= 1502.2) {
			hrecord.pressure -= 1000;
		}

		hrecord.indoorHumidity = (tempint - (tempint % 10000)) / 10000;
		hrecord.outdoorHumidity = (data[5] >> 4) * 10 + (data[5] & 0xF);
		hrecord.rainfall = ((data[7] & 0xF) * 256 + data[6]) * 0.518;
		hrecord.windspeed = (data[8] * 16 + (data[7] >> 4)) / 10.0;
		hrecord.winddir = (data[9] & 0xF) * 22.5;

		// Temperatures in Celcius. Cannot convert until WC is calculated
		tempint = ((data[2] & 0xF) << 16) + (data[1] << 8) + data[0];
		hrecord.indoorTemperature = (tempint % 1000) / 10.0 - 30.0;
		hrecord.outdoorTemperature = (tempint - (tempint % 1000)) / 10000.0 - 30.0;

		// Calculate windchill using new post 2001 USA/Canadian formula
		// Twc = 13.112 + 0.6215*Ta -11.37*V^0.16 + 0.3965*Ta*V^0.16 [Celcius
		// and km/h]
		/*
		 * wind_kmph = 3.6 * *windspeed; if (wind_kmph > 4.8) { windchill =
		 * 13.12 + 0.6215 * *temperature_outdoor - 11.37 * pow(wind_kmph, 0.16) +
		 * 0.3965 * *temperature_outdoor * pow(wind_kmph, 0.16); } else {
		 * windchill = *temperature_outdoor; }
		 */

		if (--record == -1) {
			record = 174;
		}
		hrecord.nextRecord = record;

		return hrecord;
	}

	// why are these synchronized?
	/**
	 * Returns a temperature object. 
	 * 
	 * @return TemperatureHumidity object
	 */
	public synchronized TemperatureHumidity getTemperature() {
		return temperature;
	}
	
	public synchronized Wind getWind() {
		return wind;
	}
	
	public synchronized Rain getRain() {
		return rain;
	}
	
	public synchronized Pressure getPressure() {
		return pressure;
	}
}
